"""Types."""

# pylint: disable=unused-import
from httpx import Response as HTTPResponse  # noqa: F401, pragma: no cover
from httpx._types import (  # noqa: F401, pragma: no cover
    AuthTypes,
    HeaderTypes,
    URLTypes,
)
